<?php

use MediaWiki\MediaWikiServices;
use MediaWiki\Revision\RevisionRecord;
use Wikimedia\Rdbms\IDatabase;

/**
 * Class representing a MediaWiki article and history
 *
 * FlaggableWikiPage::getTitleInstance() is preferred over constructor calls
 */
class FlaggableWikiPage extends WikiPage {
	/* Process cache variables */
	protected $stable = 0;
	protected $stableRev = null;
	protected $revsArePending = null;
	protected $pendingRevCount = null;
	protected $pageConfig = null;
	protected $syncedInTracking = null;

	protected $file = null; // for file pages

	/**
	 * Get a FlaggableWikiPage for a given title
	 * @param Title $title
	 * @return FlaggableWikiPage
	 * @suppress PhanUndeclaredProperty For Title->flaggedRevsArticle
	 */
	public static function getTitleInstance( Title $title ) {
		// Check if there is already an instance on this title
		if ( !isset( $title->flaggedRevsArticle ) ) {
			$ctitle = clone $title; // avoid cycles
			$title->flaggedRevsArticle = new self( $ctitle );
		}
		return $title->flaggedRevsArticle;
	}

	/**
	 * Transfer the prepared edit cache from a WikiPage object
	 *
	 * @param WikiPage $page
	 * @return void
	 */
	public function preloadPreparedEdit( WikiPage $page ) {
		$this->mPreparedEdit = $page->mPreparedEdit;
	}

	/**
	 * Clear object process cache values
	 * @return void
	 */
	public function clear() {
		$this->stable = 0;
		$this->stableRev = null;
		$this->revsArePending = null;
		$this->pendingRevCount = null;
		$this->pageConfig = null;
		$this->syncedInTracking = null;
		$this->file = null;
		parent::clear(); // call super!
	}

	/**
	 * Get the current file version (null if this not a File page)
	 * @return File|null|bool false
	 */
	public function getFile() {
		if ( $this->file === null && $this->mTitle->getNamespace() == NS_FILE ) {
			$this->file = MediaWikiServices::getInstance()->getRepoGroup()
				->findFile( $this->mTitle );
		}
		return $this->file;
	}

	/**
	 * Is the stable version shown by default for this page?
	 * @return bool
	 */
	public function isStableShownByDefault() {
		if ( !$this->isReviewable() ) {
			return false; // no stable versions can exist
		}
		$config = $this->getStabilitySettings(); // page configuration
		return (bool)$config['override'];
	}

	/**
	 * Do edits have to be reviewed before being shown by default (going live)?
	 * @return bool
	 */
	public function editsRequireReview() {
		return (
			$this->isReviewable() && // reviewable page
			$this->isStableShownByDefault() && // and stable versions override
			$this->getStableRev() // and there is a stable version
		);
	}

	/**
	 * Has data for this page been loaded?
	 * @return bool
	 */
	public function isDataLoaded() {
		return $this->mDataLoaded;
	}

	/**
	 * Are edits to this page currently pending?
	 * @return bool
	 */
	public function revsArePending() {
		if ( !$this->mDataLoaded ) {
			$this->loadPageData();
		}
		return $this->revsArePending;
	}

	/**
	 * Get number of revs since the stable revision
	 * Note: slower than revsArePending()
	 * @param int $flags FR_MASTER (be sure to use loadFromDB( FR_MASTER ) if set)
	 * @return int
	 */
	public function getPendingRevCount( $flags = 0 ) {
		global $wgParserCacheExpireTime;

		if ( !$this->mDataLoaded ) {
			$this->loadPageData();
		}
		# Pending count deferred even after page data load
		if ( $this->pendingRevCount !== null ) {
			return $this->pendingRevCount; // use process cache
		}
		$srev = $this->getStableRev();
		if ( !$srev ) {
			return 0; // none
		}
		$count = null;
		$sRevId = $srev->getRevId();

		$fname = __METHOD__;
		$callback = function (
			$oldValue = null, &$ttl = null, array &$setOpts = []
		) use ( $flags, $srev, $fname ) {
			$db = wfGetDB( ( $flags & FR_MASTER ) ? DB_MASTER : DB_REPLICA );
			$setOpts += Database::getCacheSetOptions( $db );

			return (int)$db->selectField(
				'revision',
				'COUNT(*)',
				[
					'rev_page' => $this->getId(),
					// T17515
					'rev_timestamp > ' .
						$db->addQuotes( $db->timestamp( $srev->getRevTimestamp() ) )
				],
				$fname
			);
		};

		if ( ( $flags & FR_MASTER ) ) {
			$this->pendingRevCount = $callback();
		} else {
			$cache = MediaWikiServices::getInstance()->getMainWANObjectCache();
			$this->pendingRevCount = (int)$cache->getWithSetCallback(
				# Confirm that cache value was made against the same stable rev Id.
				# This avoids lengthy cache pollution if $sRevId is outdated.
				$cache->makeKey( 'flaggedrevs-countPending', $this->getId(), $sRevId ),
				$wgParserCacheExpireTime,
				$callback,
				[
					'touchedCallback' => function () {
						return wfTimestampOrNull( TS_UNIX, $this->getTouched() );
					}
				]
			);
		}

		return $this->pendingRevCount;
	}

	/**
	 * Checks if the stable version is synced with the current revision
	 * Note: slower than getPendingRevCount()
	 * @return bool
	 */
	public function stableVersionIsSynced() {
		global $wgParserCacheExpireTime;

		$srev = $this->getStableRev();
		if ( !$srev ) {
			return true;
		}
		# Stable text revision must be the same as the current
		if ( $this->revsArePending() ) {
			return false;
		# Stable file revision must be the same as the current
		} elseif ( $this->mTitle->getNamespace() == NS_FILE ) {
			$file = $this->getFile(); // current upload version
			if ( $file && $file->getTimestamp() > $srev->getFileTimestamp() ) {
				return false;
			}
		}
		# If using the current version of includes, there is nothing else to check.
		if ( FlaggedRevs::inclusionSetting() == FR_INCLUDES_CURRENT ) {
			return true; // short-circuit
		}

		$cache = MediaWikiServices::getInstance()->getMainWANObjectCache();

		return (bool)$cache->getWithSetCallback(
			$cache->makeKey( 'flaggedrevs-includes-synced', $this->getId() ),
			$wgParserCacheExpireTime,
			function () use ( $srev ) {
				# Since the stable and current revisions have the same text and only outputs, the
				# only other things to check for are template and file differences in the output.
				# (a) Check if the current output has a newer template/file used
				# (b) Check if the stable version has a file/template that was deleted
				$synced = (
					!$srev->findPendingTemplateChanges() &&
					!$srev->findPendingFileChanges( 'noForeign' )
				);

				return $synced ? 1 : 0;
			},
			[
				'touchedCallback' => function () {
					return $this->getTouched();
				}
			]
		);
	}

	/**
	 * Are template/file changes and ONLY template/file changes pending?
	 * @return bool
	 */
	public function onlyTemplatesOrFilesPending() {
		return ( !$this->revsArePending() && !$this->stableVersionIsSynced() );
	}

	/**
	 * Is this page less open than the site defaults?
	 * @return bool
	 */
	public function isPageLocked() {
		return ( !FlaggedRevs::isStableShownByDefault() && $this->isStableShownByDefault() );
	}

	/**
	 * Is this page more open than the site defaults?
	 * @return bool
	 */
	public function isPageUnlocked() {
		return ( FlaggedRevs::isStableShownByDefault() && !$this->isStableShownByDefault() );
	}

	/**
	 * Tags are only shown for unreviewed content and this page is not locked/unlocked?
	 * @return bool
	 */
	public function lowProfileUI() {
		return FlaggedRevs::lowProfileUI() &&
			FlaggedRevs::isStableShownByDefault() == $this->isStableShownByDefault();
	}

	/**
	 * Is this article reviewable?
	 * @return bool
	 */
	public function isReviewable() {
		if ( !FlaggedRevs::inReviewNamespace( $this->mTitle ) ) {
			return false;
		}
		# Check if flagging is disabled for this page via config
		if ( FlaggedRevs::useOnlyIfProtected() ) {
			$config = $this->getStabilitySettings(); // page configuration
			return (bool)$config['override']; // stable is default or flagging disabled
		}
		return true;
	}

	/**
	 * Get the stable revision ID
	 * @return int
	 */
	public function getStable() {
		if ( !FlaggedRevs::inReviewNamespace( $this->mTitle ) ) {
			return 0; // short-circuit
		}
		if ( !$this->mDataLoaded ) {
			$this->loadPageData();
		}
		return (int)$this->stable;
	}

	/**
	 * Get the stable revision
	 * @return FlaggedRevision|null
	 */
	public function getStableRev() {
		if ( !FlaggedRevs::inReviewNamespace( $this->mTitle ) ) {
			return null; // short-circuit
		}
		if ( !$this->mDataLoaded ) {
			$this->loadPageData();
		}
		# Stable rev deferred even after page data load
		if ( $this->stableRev === null ) {
			$srev = FlaggedRevision::newFromTitle( $this->mTitle, $this->stable );
			$this->stableRev = $srev ?: false; // cache negative hits too
		}
		return $this->stableRev ?: null; // false => null
	}

	/**
	 * Get visibility restrictions on page
	 * @return array (select,override)
	 */
	public function getStabilitySettings() {
		if ( !$this->mDataLoaded ) {
			$this->loadPageData();
		}
		return $this->pageConfig;
	}

	/**
	 * Get the fp_reviewed value for this page
	 * @return bool
	 */
	public function syncedInTracking() {
		if ( !$this->mDataLoaded ) {
			$this->loadPageData();
		}
		return $this->syncedInTracking;
	}

	/**
	 * Get the newest of the highest rated flagged revisions of this page
	 * Note: will not return deleted revisions
	 * @return int
	 */
	public function getBestFlaggedRevId() {
		$dbr = wfGetDB( DB_REPLICA );
		# Get the highest quality revision (not necessarily this one).
		$oldid = $dbr->selectField( [ 'flaggedrevs', 'revision' ],
			'fr_rev_id',
			[
				'fr_page_id' => $this->getId(),
				'rev_page = fr_page_id', // sanity
				'rev_id = fr_rev_id',
				$dbr->bitAnd( 'rev_deleted', RevisionRecord::DELETED_TEXT ) . ' = 0'
			],
			__METHOD__,
			[
				'ORDER BY' 	=> 'fr_quality DESC, fr_rev_timestamp DESC'
			]
		);
		return (int)$oldid;
	}

	/**
	 * Updates the fp_reviewed field for this article
	 */
	public function lazyUpdateSyncStatus() {
		if ( wfReadOnly() ) {
			return;
		}

		JobQueueGroup::singleton()->push(
			new FRExtraCacheUpdateJob(
				$this->getTitle(),
				[ 'type' => 'updatesyncstate' ]
			)
		);
	}

	/**
	 * Fetch a page record with the given conditions
	 * @param IDatabase $dbr Database object
	 * @param array $conditions
	 * @param array $options
	 * @return mixed Database result resource, or false on failure
	 */
	protected function pageData( $dbr, $conditions, $options = [] ) {
		$pageQuery = WikiPage::getQueryInfo();
		$row = $dbr->selectRow(
			array_merge( $pageQuery['tables'], [ 'flaggedpages', 'flaggedpage_config' ] ),
			array_merge(
				$pageQuery['fields'],
				FRPageConfig::selectFields(),
				[ 'fp_pending_since', 'fp_stable', 'fp_reviewed' ]
			),
			$conditions,
			__METHOD__,
			$options,
			$pageQuery['joins'] + [
				'flaggedpages' 		 => [ 'LEFT JOIN', 'fp_page_id = page_id' ],
				'flaggedpage_config' => [ 'LEFT JOIN', 'fpc_page_id = page_id' ],
			]
		);
		return $row;
	}

	/**
	 * Set the page field data loaded from some source
	 * @param stdClass|string|int $data Database row object or "fromdb" or "fromdbmaster"
	 * @return void
	 */
	public function loadPageData( $data = self::READ_NORMAL ) {
		$this->mDataLoaded = true; // sanity
		# Fetch data from DB as needed...
		$from = WikiPage::convertSelectType( $data );
		if ( $from === self::READ_NORMAL || $from === self::READ_LATEST ) {
			$db = wfGetDB( $from === self::READ_LATEST ? DB_MASTER : DB_REPLICA );
			$data = $this->pageDataFromTitle( $db, $this->mTitle );
		}
		# Load in primary page data...
		parent::loadPageData( $data /* Row obj */ );
		# Load in FlaggedRevs page data...
		$this->stable = 0; // 0 => "found nothing"
		$this->stableRev = null; // defer this one...
		$this->revsArePending = false; // false => "found nothing" or "none pending"
		$this->pendingRevCount = null; // defer this one...
		$this->pageConfig = FRPageConfig::getDefaultVisibilitySettings(); // default
		$this->syncedInTracking = true; // false => "unreviewed" or "synced"
		# Load in flaggedrevs Row data if the page exists...(sanity check NS)
		if ( $data && FlaggedRevs::inReviewNamespace( $this->mTitle ) ) {
			if ( $data->fpc_override !== null ) { // page config row found
				$this->pageConfig = FRPageConfig::getVisibilitySettingsFromRow( $data );
			}
			if ( $data->fp_stable !== null ) { // stable rev found
				$this->stable = (int)$data->fp_stable;
				$this->revsArePending = ( $data->fp_pending_since !== null ); // revs await review
				$this->syncedInTracking = (bool)$data->fp_reviewed;
			}
		}
	}

	/**
	 * Updates the flagging tracking tables for this page
	 * @param FlaggedRevision $srev The new stable version
	 * @param int|null $latest The latest rev ID (optional)
	 * @return bool Updates were done
	 */
	public function updateStableVersion( FlaggedRevision $srev, $latest = null ) {
		if ( !$this->exists() ) {
			return false; // no bogus entries
		}

		$revRecord = $srev->getRevisionRecord();
		if ( !$revRecord ) {
			return false; // no bogus entries
		}

		# Get the latest revision ID if not set
		if ( !$latest ) {
			$latest = $this->mTitle->getLatestRevID( Title::GAID_FOR_UPDATE );
		}
		$dbw = wfGetDB( DB_MASTER );
		# Get the highest quality revision (not necessarily this one)...
		if ( $srev->getQuality() === FlaggedRevs::highestReviewTier() ) {
			$maxQuality = $srev->getQuality(); // save a query
		} else {
			$maxQuality = $dbw->selectField( [ 'flaggedrevs', 'revision' ],
				'fr_quality',
				[ 'fr_page_id' => $this->getId(),
					'rev_id = fr_rev_id',
					'rev_page = fr_page_id',
					$dbw->bitAnd( 'rev_deleted', RevisionRecord::DELETED_TEXT ) . ' = 0'
				],
				__METHOD__,
				[ 'ORDER BY' => 'fr_quality DESC', 'LIMIT' => 1 ]
			);
			$maxQuality = max( $maxQuality, $srev->getQuality() ); // sanity
		}
		# Get the timestamp of the first edit after the stable version (if any)...
		$nextTimestamp = null;
		if ( $revRecord->getId() != $latest ) {
			$timestamp = $dbw->timestamp( $revRecord->getTimestamp() );
			$nextEditTS = $dbw->selectField( 'revision',
				'rev_timestamp',
				[
					'rev_page' => $this->getId(),
					"rev_timestamp > " . $dbw->addQuotes( $timestamp ) ],
				__METHOD__,
				[ 'ORDER BY' => 'rev_timestamp ASC', 'LIMIT' => 1 ]
			);
			if ( $nextEditTS ) { // sanity check
				$nextTimestamp = $nextEditTS;
			}
		}
		# Get the new page sync status...
		$synced = !(
			$nextTimestamp !== null || // edits pending
			$srev->findPendingTemplateChanges() || // template changes pending
			$srev->findPendingFileChanges( 'noForeign' ) // file changes pending
		);
		# Alter table metadata
		$dbw->replace(
			'flaggedpages',
			'fp_page_id',
			[
				'fp_page_id'       => $revRecord->getPageId(), // Don't use $this->getId(), T246720
				'fp_stable'        => $revRecord->getId(),
				'fp_reviewed'      => $synced ? 1 : 0,
				'fp_quality'       => ( $maxQuality === false ) ? null : $maxQuality,
				'fp_pending_since' => $dbw->timestampOrNull( $nextTimestamp )
			],
			__METHOD__
		);
		# Update pending edit tracking table
		self::updatePendingList( $this->getId(), $latest );
		return true;
	}

	/**
	 * Updates the flagging tracking tables for this page
	 */
	public function clearStableVersion() {
		if ( !$this->exists() ) {
			return; // nothing to do
		}
		$dbw = wfGetDB( DB_MASTER );
		$dbw->delete( 'flaggedpages', [ 'fp_page_id' => $this->getId() ], __METHOD__ );
		$dbw->delete( 'flaggedpage_pending', [ 'fpp_page_id' => $this->getId() ], __METHOD__ );
	}

	/**
	 * Updates the flaggedpage_pending table
	 * @param int $pageId Page ID
	 * @param int $latest Latest revision
	 */
	protected static function updatePendingList( $pageId, $latest ) {
		$data = [];
		# Get the highest tier used on this wiki
		$level = FlaggedRevs::highestReviewTier();

		$dbw = wfGetDB( DB_MASTER );
		# Update pending times for each level, going from highest to lowest
		$higherLevelId = 0;
		$higherLevelTS = '';
		while ( $level >= 0 ) {
			# Get the latest revision of this level...
			# Any revision of one tier is also a revision of lower tiers.
			# Instead of doing fr_quality > X queries we do exact comparisons
			# for better INDEX usage. However, in order to treat a rev as the
			# latest tier X rev, we make sure it is newer than all tier (X+1) revs.
			$row = $dbw->selectRow(
				[ 'flaggedrevs', 'revision' ],
				[ 'fr_rev_id', 'rev_timestamp' ],
				[
					'fr_page_id' => $pageId,
					'fr_quality' => $level, // this level
					'fr_rev_timestamp > ' . $dbw->addQuotes( $higherLevelTS ),
					'rev_id = fr_rev_id', // rev exists
					'rev_page = fr_page_id', // sanity
					$dbw->bitAnd( 'rev_deleted', RevisionRecord::DELETED_TEXT ) . ' = 0'
				],
				__METHOD__,
				[ 'ORDER BY' => 'fr_rev_timestamp DESC', 'LIMIT' => 1 ]
			);
			# If there is a revision of this level, track it...
			# Revisions accepted to one tier count as accepted
			# at the lower tiers (i.e. quality -> checked).
			if ( $row ) {
				$id = $row->fr_rev_id;
				$ts = $row->rev_timestamp;
			} else { // use previous rev of higher tier (if any)
				$id = $higherLevelId;
				$ts = $higherLevelTS;
			}
			# Get edits that actually are pending...
			if ( $id && $latest > $id ) {
				# Get the timestamp of the edit after this version (if any)
				$nextTimestamp = $dbw->selectField( 'revision',
					'rev_timestamp',
					[ 'rev_page' => $pageId, "rev_timestamp > " . $dbw->addQuotes( $ts ) ],
					__METHOD__,
					[ 'ORDER BY' => 'rev_timestamp ASC', 'LIMIT' => 1 ]
				);
				$data[] = [
					'fpp_page_id'       => $pageId,
					'fpp_quality'       => $level,
					'fpp_rev_id'        => $id,
					'fpp_pending_since' => $nextTimestamp
				];
				$higherLevelId = $id;
				$higherLevelTS = $ts;
			}
			$level--;
		}
		# Clear any old junk, and insert new rows
		$dbw->delete( 'flaggedpage_pending', [ 'fpp_page_id' => $pageId ], __METHOD__ );
		$dbw->insert( 'flaggedpage_pending', $data, __METHOD__ );
	}
}
